local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local SM = LibStub("LibSharedMedia-3.0")
local TalentQuery = LibStub("LibTalentQuery-1.0")
local commPrefix = "Utopia"
local db, engine
local playerClass
local new, del, copy, deepDel, Gradient, UnitFullName, ClassColour, SmoothColour, rotate
local title

local module = LibStub("AceAddon-3.0"):NewAddon("Utopia", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
Utopia = module
module.version = tonumber(strmatch("$Revision: 34 $", "(%d+)")) or 0
function Utopia:UpdateVersion(ver)
	self.version = max(tonumber(strmatch(ver, "(%d+)")) or 0, self.version)
end

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Utopia", {
	type = "data source",
	label = L["TITLE"],
	icon = select(3, GetSpellInfo(47467)),
})

-- LDB Bits
function LDB:Update()
	self.label = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
end

-- OnClick
function LDB.OnClick(self, button)
	if (not IsModifierKeyDown()) then
		if (button == "LeftButton") then
			module:ToggleActive()
		elseif (button == "RightButton") then
			module:OpenConfig()
		end
	end
end

-- OnTooltipShow
function LDB.OnTooltipShow(tt)
	tt:AddLine(module.title)
end

module.options = {
	type = 'group',
	order = 15,
	name = L["TITLE"],
	desc = L["Configuration"],
	handler = module,
	args = {
		General = {
			type = 'group',
			order = 1,
			name = L["General Settings"],
			desc = L["General Settings"],
			args = {
				enable = {
					type = 'toggle',
					name = L["Enable"],
					desc = L["Enable target debuff status display"],
					get = function() return db.enable end,
					set = function(info, n) db.enable = n module:ShowWhen() end,
					order = 5,
				},
				locked = {
					type = 'toggle',
					name = L["Locked"],
					desc = L["Toggle locked position. When unlocked, you can move the display with the bar"],
					get = function() return db.locked end,
					set = function(info, n)
						db.locked = n
						module.frame.bar:EnableMouse(not db.locked)
		
						for i,icon in pairs(module.frame.icons) do
							if (n) then
								icon:RegisterForDrag(nil)
							else
								icon:RegisterForDrag("LeftButton")
							end
						end

						if (module.frame.backdrop) then
							module.frame.backdrop:EnableMouse(not db.locked)
						end
					end,
					order = 20,
				},
				display = {
					type = 'group',
					order = 100,
					name = L["Display"],
					desc = L["Display options for the debuff monitor"],
					guiInline = true,
					args = {
						orientation = {
							type = 'select',
							name = L["Orientation"],
							desc = L["Set the orientation for the display"],
							values = {V = L["Vertical"], H = L["Horizontal"]},
							order = 5,
							get = function() return db.orientation end,
							set = "SetOrientation",
						},
						side = {
							type = 'select',
							name = L["Side"],
							desc = L["Which side to put the bar"],
							values = {B = L["Bottom/Right"], T = L["Top/Left"]},
							order = 10,
							get = function() return db.side end,
							set = "SetOrientation",
						},
						size = {
							type = "range",
							name = L["Icon Size"],
							desc = L["Set the size of the icons"],
							func = timeFunc,
							order = 2,
							get = function() return db.size end,
							set = "SetIconSize",
							min = 8,
							max = 64,
							step = 1,
							order = 20,
						},
						scale = {
							type = "range",
							name = L["Scale"],
							desc = L["Set the scale of the icons"],
							func = timeFunc,
							order = 2,
							get = function() return db.scale end,
							set = function(info, n) db.scale = n module:SetScale() module:SavePosition(module.frame) end,
							min = 0.2,
							max = 2,
							step = 0.01,
							bigStep = 0.1,
							order = 30,
						},
						texture = {
							type = 'select', dialogControl = 'LSM30_Statusbar',
							name = L["Bar Texture"],
							desc = L["Set the texture for the bar"],
							values = AceGUIWidgetLSMlists.statusbar,
							order = 40,
							get = function() return db.texture end,
							set = function(v,n)
								db.texture = n
								local tex = SM:Fetch("statusbar", n)
								module.frame.bar:SetStatusBarTexture(tex)
								module.frame.bar.bg:SetTexture(tex)
							end,
						},
						barwidth = {
							type = "range",
							name = L["Width"],
							desc = L["Set the width of the status bar"],
							func = timeFunc,
							order = 2,
							get = function() return db.barwidth end,
							set = function(info, n) db.barwidth = n module:SetOrientation() end,
							min = 1,
							max = 20,
							step = 1,
							order = 50,
						},
					},
				},
				buttons = {
					type = 'group',
					name = L["Buttons"],
					order = 200,
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value)
						db[info[#info]] = value
						module:SetOrientation()
					end,
					args = {
						hspacing = {
							type = "range",
							name = L["Horizontal Spacing"],
							desc = L["Set the horizontal spacing of the buttons"],
							func = timeFunc,
							order = 2,
							min = 0,
							max = 20,
							step = 1,
							order = 1,
						},
						vspacing = {
							type = "range",
							name = L["Vertical Spacing"],
							desc = L["Set the vertical spacing of the buttons"],
							func = timeFunc,
							order = 2,
							min = 0,
							max = 20,
							step = 1,
							order = 2,
						},
						columns = {
							type = "range",
							name = L["Columns"],
							desc = L["Set the buttons before breaking to next row/column"],
							func = timeFunc,
							order = 2,
							min = 1,
							max = 15,
							step = 1,
							order = 3,
						},
					},
				},
				colours = {
					type = 'group',
					name = L["Colour"],
					order = 300,
					guiInline = true,
					get = function(info)
						local c = db.colour[info[#info]]
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)		--value)
						local c = db.colour[info[#info]]
						c.r, c.g, c.b, c.a = r, g, b, a
						if (module.frame) then
							module.frame:UpdateBuffs()
						end
					end,
					args = {
						active = {
							type = "color",
							name = L["Active Colour"],
							desc = L["Set the colour for debuffs which your group has applied."],
							hasAlpha = true,
							order = 1,
						},
						inactive = {
							type = "color",
							name = L["Inactive Colour"],
							desc = L["Set the colour for debuffs which your group is able to apply, but has not done so."],
							hasAlpha = true,
							order = 2,
						},
          				unavailable = {
							type = "color",
							name = L["Unavailable Colour"],
							desc = L["Set the colour for debuffs which your group is not able to apply."],
							hasAlpha = true,
							order = 3,
						},
					},
				},
				background = {
					type = "group",
					name = L["Background Options"],
					guiInline = true,
					order = 400,
					get = function(info) return db.Background[ info[#info] ] end,
					set = function(info, value)
						db.Background[ info[#info] ] = value
						module:UpdateBackdrop()
					end,
					args = {
						Enable = {
							type = "toggle",
							order = 1,
							name = L["Enable"],
							desc = L["Enable the frame background"],
						},
						Texture = {
							type = "select", dialogControl = 'LSM30_Background',
							order = 10,
							name = L["Background Texture"],
							desc = L["Texture to use for the frame's background"],
							values = AceGUIWidgetLSMlists.background,
						},
						BorderTexture = {
							type = "select", dialogControl = 'LSM30_Border',
							order = 20,
							name = L["Border Texture"],
							desc = L["Texture to use for the frame's border"],
							values = AceGUIWidgetLSMlists.border,
						},
						Color = {
							type = "color",
							order = 30,
							name = L["Background Color"],
							desc = L["Frame's background color"],
							hasAlpha = true,
							get = function(info)
								local t = db.Background.Color
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = db.Background.Color
								t.r, t.g, t.b, t.a = r, g, b, a
								module:UpdateBackdrop()
							end,
						},
						BorderColor = {
							type = "color",
							order = 40,
							name = L["Border Color"],
							desc = L["Frame's border color"],
							hasAlpha = true,
							get = function(info)
								local t = db.Background.BorderColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = db.Background.BorderColor
								t.r, t.g, t.b, t.a = r, g, b, a
								module:UpdateBackdrop()
							end,
						},
						Tile = {
							type = "toggle",
							order = 50,
							name = L["Tile Background"],
							desc = L["Tile the background texture"],
						},
						TileSize = {
							type = "range",
							order = 60,
							name = L["Background Tile Size"],
							desc = L["The size used to tile the background texture"],
							min = 16, max = 256, step = 1,
							disabled = function() return not db.Background.Tile end,
						},
						EdgeSize = {
							type = "range",
							order = 70,
							name = L["Border Thickness"],
							desc = L["The thickness of the border"],
							min = 1, max = 24, step = 1,
						},
					},
				},
			},
		},
		show = {
			type = 'group',
			name = L["Show When"],
			desc = L["Set rules for when to show Utopia"],
			get = function(info) return db.show[info[#info]] end,
			set = function(info, value)
				db.show[info[#info]] = value
				module:ShowWhen()
			end,
			order = 50,
			args = {
				always = {
					type = 'toggle',
					name = L["Always Show"],
					desc = L["Always show the bar"],
					order = 1,
				},
				solo = {
					type = 'toggle',
					name = SOLO,
					desc = L["Show when solo"],
					order = 10,
				},
				party = {
					type = 'toggle',
					name = PARTY,
					desc = L["Show when in a party"],
					order = 20,
				},
				raid = {
					type = 'toggle',
					name = RAID,
					desc = L["Show when in a raid"],
					order = 30,
				},
				boss = {
					type = 'toggle',
					name = L["Bosses"],
					desc = L["Only show when targetting boss mobs"],
					order = 40,
				},
			},
		},
	},
}

AceConfig:RegisterOptionsTable("Utopia", module.options)

-- InitData
function module:InitData()
	del(self.lookup)
	self.lookup = new()
	deepDel(self.classes)
	self.classes = new()

	for name,info in pairs(self.data) do
		info.key = name
		for spell,spellInfo in pairs(info.spells) do
			spellInfo.name = spell
			spellInfo.key = name
			self.lookup[spell] = spellInfo

			local c = self.classes[spellInfo.class]
			if (not c) then
				c = new()
				self.classes[spellInfo.class] = c
			end
			c[name] = spell
		end
	end

	del(self.mappingLookup)
	self.mappingLookup = new()

	for targetSpell,data in pairs(self.mapping) do
		data.key = targetSpell
		for sourceSpell,info in pairs(data.list) do
			self.mappingLookup[sourceSpell] = data
		end
	end
end

-- SavePosition
function module:SavePosition(frame)
	local point, _, relPoint, x, y = frame:GetPoint(1)
	self.db.char.pos[frame:GetName()] = {point, "UIParent", relPoint, x, y}
end

-- RestorePosition
function module:RestorePosition(frame)
	if (type(self.db.char.pos) == "table") then
		-- Convert old pos to new pos savedvars
		if (#self.db.char.pos == 5 and not self.db.char.pos.UtopiaAnchorFrame) then
			local old = self.db.char.pos
			self.db.char.pos = new()
			self.db.char.pos.UtopiaAnchorFrame = old
		end
	end

	local pos = self.db.char.pos and self.db.char.pos[frame:GetName()]
	if (pos) then
		frame:SetPoint(unpack(pos))
	else
		frame:SetPoint("CENTER", UIParent, "TOP", 0, -150)
	end
end

-- myLevelRank
local function myLevelRank(ranks)
	if (ranks) then
		local lvl = UnitLevel("player")
		local rank
		for i,level in pairs(ranks) do
			if (lvl >= level) then
				rank = i
			end
		end
		return rank
	end
end

-- OnDebuffApplied
function module:OnDebuffApplied(dstGUID, spellName, debuff, dose)
	if (UnitGUID(self.frame.unit) == dstGUID) then
		self.frame:OnDebuffApplied(spellName, debuff, dose)
	end
end

-- OnDebuffRemoved
function module:OnDebuffRemoved(dstGUID, spellName, debuff, dose)
	if (UnitGUID(self.frame.unit) == dstGUID) then
		self.frame:OnDebuffRemoved(spellName, debuff, dose)
	end
end

-- getAmount
local function getAmount(info, short, debuff, player, getMaximum)
	local ret = info.amount
	local suffix

	local name, rank, stacks
	if (debuff) then
		if (debuff.rank) then
			rank = debuff.rank
		else
			name, rank = GetSpellInfo(debuff.id)
			assert(name == debuff.name)
			if (rank) then
				rank = tonumber(rank:match("(%d+)"))
			end
		end
		if (info.maxStacks) then
			stacks = debuff.stacks
		elseif (debuff.stacks and debuff.stacks > 1) then
			error(format("%s has %d stacks. info.maxStacks not set", info.name, stacks))
		end
	else
		rank = myLevelRank(info.rankLevels)
	end

	local playerName = player and player.name

	if (info.amountsPerStack) then
		ret = info.amountsPerStack[rank or #info.amountsPerStack]
		assert(ret)
		if (not short) then
			suffix = format(L["(for %d %s)"], stacks or info.maxStacks, L["stacks"])
		end
	elseif (info.amountPerStack) then
		ret = info.amountPerStack
		assert(ret)
		if (not short) then
			suffix = format(L["(for %d %s)"], stacks or info.maxStacks, L["stacks"])
		end
	elseif (info.amounts) then
		ret = info.amounts[rank or #info.amounts]
		assert(ret)
	end

	if (info.improvedAmountPerTalentPoint) then
		if (playerName) then
			local points = module:HasTalent(playerName, info.improved)
			if (points) then
--@debug@
				if (points > info.maxTalentPoints) then
					error(format("%s's points more than maxTalentPoints for %s", player.name, info.improved))
				end
--@no-debug@
				ret = ret + info.improvedAmountPerTalentPoint * points
			end
		else
			local potentialLevel = not getMaximum and module.spellPotential and module.spellPotential[info.name]
			if (potentialLevel) then
--@debug@
				assert(not info.maxTalentPoints or potentialLevel <= info.maxTalentPoints, "potentialLevel <= info.maxTalentPoints")
--@no-debug@
				if (potentialLevel > 0) then
					ret = ret + info.improvedAmountPerTalentPoint * potentialLevel
				end
			else
				ret = ret + info.improvedAmountPerTalentPoint * (info.maxTalentPoints or 1)
			end
		end

	elseif (info.improvedPercentagePerTalentPoint) then
		if (playerName) then
			local points = module:HasTalent(playerName, info.improved)
			if (points) then
--@debug@
				if (points > info.maxTalentPoints) then
					error(format("%s's points(%d) more than maxTalentPoints(%d) for %s", player.name, points, info.maxTalentPoints, info.improved))
				end
--@no-debug@
				ret = ret * (1 + (info.improvedPercentagePerTalentPoint * points / 100))
			end
		else
			local potentialLevel = not getMaximum and module.spellPotential and module.spellPotential[info.name]
			if (potentialLevel) then
--@debug@
				assert(not info.maxTalentPoints or potentialLevel <= info.maxTalentPoints, "potentialLevel <= info.maxTalentPoints")
--@no-debug@
				if (potentialLevel > 0) then
					ret = ret * (1 + (info.improvedPercentagePerTalentPoint * potentialLevel / 100))
				end
			else
				ret = ret * (1 + (info.improvedPercentagePerTalentPoint * (info.maxTalentPoints or 1) / 100))
			end
		end

	elseif (info.amountPerTalentPoint) then
		if (playerName) then
			local points = module:HasTalent(playerName, info.requiredTalent)
			if (points) then
--@debug@
				if (points > info.maxTalentPoints) then
					error(format("%s's points(%d) more than maxTalentPoints(%d) for %s", player.name, points, info.maxTalentPoints, info.improved))
				end
				if (rank and rank ~= points) then
					error(format("%s's points(%d) assumes spell rank(%d) is the same for %s", player.name, points, rank, info.improved))
				end
--@no-debug@
				ret = info.amountPerTalentPoint * points
			end
		else
			local potentialLevel = not getMaximum and module.spellPotential and module.spellPotential[info.name]
			if (potentialLevel) then
--@debug@
				assert(not info.maxTalentPoints or potentialLevel <= info.maxTalentPoints, "potentialLevel <= info.maxTalentPoints")
--@no-debug@
				if (potentialLevel > 0) then
					ret = info.amountPerTalentPoint * potentialLevel
				end
			else
				ret = info.amountPerTalentPoint * (info.maxTalentPoints or 1)
			end
		end
	end

	if (info.maxStacks) then
		if (info.assumeMaxStacks) then
			ret = ret * info.maxStacks
		else
			ret = ret * (stacks or info.maxStacks or 1)
		end
		if (not short) then
			suffix = format(L["(for %d %s)"], (info.assumeMaxStacks and info.maxStacks) or stacks or info.maxStacks, info.stacksDescription or L["stacks"])
		end
--@debug@
	else
		if (stacks and stacks > 1 and not info.maxStacks) then
			error(format("%s has %d stacks. info.maxStacks not set", info.name, stacks))
		end
--@no-debug@
	end

	if (ret) then
		ret = floor(0.5 + ret)
		if (info.amountType) then
			ret = ret..info.amountType
		end
		if (suffix) then
			return format("%s %s", ret, suffix)
		end
	end
	return ret or ""
end
Utopia.getAmount = getAmount

local function dragStart()
	module.frame:StartMoving()
	GameTooltip:Hide()
end
local function dragStop()
	module.frame:StopMovingOrSizing()
	module:SavePosition(module.frame)
end

-- InitFrames
function module:InitFrames()
	local frame = CreateFrame("Frame", "UtopiaAnchorFrame", UIParent)
	self.frame = frame
	frame:SetHeight(1)
	frame:SetWidth(1)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame.unit = "target"

	self:SetScale()
	self:RestorePosition(self.frame)

	frame.icons = {}

	local list = new()
	for key,data in pairs(self.data) do
		tinsert(list, key)
	end
	sort(list)

	local info
	local function sortList(a, b)
		local aa = info.spells[a]
		local bb = info.spells[b]
		aa = getAmount(aa, true)
		bb = getAmount(bb, true)
		
		if (aa == bb) then
			return a < b
		else
			return aa > bb
		end
	end

	local function onEnter(self)
		GameTooltip:SetOwner(self, "LEFT")
		GameTooltip:SetText(self.key, 1, 1, 1)
		info = module.data[self.key]					-- DO NOT LOCALIZE, it's up there @ line -17
		local mob = engine:GetMobByUnit(self.parent.unit)

		local list = new()
		for name,spellInfo in pairs(info.spells) do
			tinsert(list, name)
		end
		sort(list, sortList)

		for i = 1,#list do
			local name = list[i]
			local spellInfo = info.spells[name]
			local debuff = mob and mob.debuffs[name]

			local canDo = module.spellPotential and module.spellPotential[name]
			if (not canDo) then
				if (spellInfo.class == "HUNTER") then
					canDo = module.petPotential and module.petPotential[name]
				end
			end
			local div = canDo and 1 or 2

			local c = RAID_CLASS_COLORS[spellInfo.class]
			local potentialAmount = getAmount(spellInfo, nil, debuff)
			GameTooltip:AddDoubleLine(spellInfo.pet and format("%s (%s)", name, spellInfo.pet) or name, potentialAmount, c.r / div, c.g / div, c.b / div, 1 / div, 1 / div, 1 / div)

			if (canDo) then
      			GameTooltip:AddTexture(debuff and "Interface\\RAIDFRAME\\ReadyCheck-Ready" or "Interface\\RaidFrame\\ReadyCheck-NotReady")
			end

			if (debuff and debuff.source) then
				local player = engine.players[debuff.source]
				if (player) then
					local pet

					if (player.owner) then
						pet = player
						player = engine.players[UnitGUID(pet.ownerid)]
					end

					local _, pclass = UnitClass(player.name)		-- Just to be sure
--@debug@
					assert(pclass == spellInfo.class, "pclass == spellInfo.class")
--@no-debug@
					local actualAmount = getAmount(spellInfo, nil, debuff, player)

					local who
					if (pet) then
						who = format(L["  by %s (%s%s's pet)"], pet.name, ClassColour(pclass), player.name)
					else
						who = format(L["  by %s%s"], ClassColour(pclass), player.name)
					end

					if (actualAmount ~= potentialAmount) then
						GameTooltip:AddDoubleLine(who, actualAmount, 0.5, 0.5, 0.5, 1 / div, 1 / div, 1 / div)
					else
						GameTooltip:AddLine(who, 0.5, 0.5, 0.5)
					end
				end
			end
		end

		GameTooltip:Show()
	end

	local function onLeave(self)
		GameTooltip:Hide()
	end

	local function iconActivate(self, debuff)
		local c = db.colour.active
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)

		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	function iconShowCount(self, count)
		if (count) then
			if (not self.count) then
				self.count = self:CreateFontString(nil, "OVERLAY", "GameFontNormal")
				self.count:SetAllPoints()
			end
			self.count:SetText(count)
				
		elseif (self.count) then
			self.count:Hide()
		end

		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function iconDeactivate(self)
		local c = db.colour.inactive
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.count) then
			self.count:Hide()
		end
		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function iconUnavailable(self)
		local c = db.colour.unavailable
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.count) then
			self.count:Hide()
		end
		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function buttonMouseUp(self, button)
		if (button == "LeftButton") then
			module:ToggleDetails(self.key)
		end
	end

	for i = 1,#list do
		local key = list[i]
		local data = self.data[key]

		local icon = CreateFrame("Frame", frame:GetName()..key, frame)
		icon:SetWidth(db.size)
		icon:SetHeight(db.size)
		icon:EnableMouse(true)

		if (not db.locked) then
			icon:RegisterForDrag("LeftButton")
		end

		icon.key = key
		icon.parent = frame

		icon.tex = icon:CreateTexture(nil, "OVERLAY")
		icon.tex:SetAllPoints()
		icon.tex:SetTexture(data.icon)

		icon:SetScript("OnEnter", onEnter)
		icon:SetScript("OnLeave", onLeave)
		icon:SetScript("OnDragStart", dragStart)
		icon:SetScript("OnDragStop", dragStop)
		icon:SetScript("OnMouseUp", buttonMouseUp)

		icon.Activate = iconActivate
		icon.Deactivate = iconDeactivate
		icon.Unavailable = iconUnavailable
		icon.ShowCount = iconShowCount

		tinsert(frame.icons, icon)
	end
	del(list)

	frame.bar = CreateFrame("StatusBar", frame:GetName().."Bar", frame)
	self:SetOrientation()
	
	frame.bar:SetStatusBarTexture(SM:Fetch("statusbar", db.texture))
	frame.bar:SetStatusBarColor(0, 1, 0)
	frame.bar:SetMinMaxValues(0, #frame.icons)

	frame.bar:RegisterForDrag("LeftButton")
	frame.bar:SetScript("OnDragStart", dragStart)
	frame.bar:SetScript("OnDragStop", dragStop)

	frame.bar.bg = frame.bar:CreateTexture(nil, "BACKGROUND")
	frame.bar.bg:SetAllPoints()
	frame.bar.bg:SetTexture(SM:Fetch("statusbar", db.texture))
	frame.bar.bg:SetVertexColor(0, 1, 0, 0.3)

	self:SetFrameSize()
	frame:Hide()

	if (not db.locked) then
		frame.bar:EnableMouse(true)
	end

	frame:SetScript("OnShow",
		function(self)
			self:UpdateBuffs()
			--self:RegisterEvent("UNIT_AURA")

			if (self.unit == "target" or self.unit == "targettarget") then
				self:RegisterEvent("PLAYER_TARGET_CHANGED")
			else
				self:RegisterEvent("PLAYER_FOCUS_CHANGED")
			end
		end
	)
	frame:SetScript("OnHide",
		function(self)
			--self:UnregisterEvent("UNIT_AURA")
			self:UnregisterEvent("PLAYER_TARGET_CHANGED")
			self:UnregisterEvent("PLAYER_FOCUS_CHANGED")
		end
	)
	frame:SetScript("OnEvent",
		function(self, event, ...)
			local f = self[event]
			if (f) then
				f(self, ...)
			end
		end
	)

	function frame:UNIT_AURA(unit)
		--if (unit == self.unit) then
		--	self:UpdateBuffs()
		--end
	end

	function frame:PLAYER_TARGET_CHANGED()
		if (not UnitCanAttack("player", "target") and UnitCanAttack("player", "targettarget")) then
			self.unit = "targettarget"
		else
			self.unit = "target"
		end
		self:UpdateBuffs()
	end
	function frame:PLAYER_FOCUS_CHANGED()
		if (not UnitCanAttack("player", "focus") and UnitCanAttack("player", "focustarget")) then
			self.unit = "focustarget"
		else
			self.unit = "focus"
		end
		self:UpdateBuffs()
	end

	-- AllOff
	function frame:AllOff()
		local totalAvailable = 0
		for i,icon in ipairs(self.icons) do
			icon.data = deepDel(icon.data)
			if (module.potential and module.potential[icon.key]) then
				totalAvailable = totalAvailable + 1
				icon:Deactivate()
			else
				if (module.petPotentialKeys and module.petPotentialKeys[icon.key]) then
					totalAvailable = totalAvailable + 1
					icon:Deactivate()
				else
					icon:Unavailable()
				end
			end
		end
		self.bar:SetMinMaxValues(0, totalAvailable)
		self.bar:SetValue(0)
		self.bar:SetStatusBarColor(1, 0, 0)
		self.bar.bg:SetVertexColor(1, 0, 0, 0.3)
	end

	-- GetMaxAvailable
	function frame:GetMaxAvailable()
		local totalAvailable = 0
		for i,icon in ipairs(self.icons) do
			if (module.potential and module.potential[icon.key]) then
				totalAvailable = totalAvailable + 1
			else
				if (module.petPotentialKeys and module.petPotentialKeys[icon.key]) then
					totalAvailable = totalAvailable + 1
				end
			end
		end
		return totalAvailable
	end

	-- GetIconForKey
	function frame:GetIconForKey(key)
		for i,icon in ipairs(self.icons) do
			if (icon.key == key) then
				return icon
			end
		end
--@debug@
		module:Print(format("No icon for key |cFF80FF80%s|r", key))
--@no-debug@
	end

	-- GetIconForSpell
	function frame:GetIconForSpell(spellName)
		local spell = module.lookup[spellName]
--@debug@
		if (not spell) then
			module:Print(format("No spell lookup for |cFF80FF80%s|r", spellName))
			return
		end
--@no-debug@

		return self:GetIconForKey(spell.key)
	end

	-- OnDebuffApplied
	function frame:OnDebuffApplied(spellName, debuff, dose)
		--module:Print(format("frame:OnDebuffApplied(%s)", spellName))

		local icon = self:GetIconForSpell(spellName)
--@debug@
		assert(icon)
--@no-debug@

		local spellInfo = module.lookup[spellName]
--@debug@
		assert(spellInfo)
		assert(spellInfo.key == icon.key)
--@no-debug@

		if (spellInfo.exclusiveWith) then
			-- Special case for paladins
			local altIcon = self:GetIconForKey(spellInfo.exclusiveWith)
			if (module.classcount[spellInfo.class] < 2) then
				self.bar:SetMinMaxValues(0, self:GetMaxAvailable() - 1)
				altIcon:Unavailable()
			else
				self.bar:SetMinMaxValues(0, self:GetMaxAvailable())
				altIcon:Deactivate()
			end
		end

		if (not icon.data) then
			icon.data = new()
		end
		if (not icon.data[spellName]) then
			local val = self.bar:GetValue() + 1
			self.bar:SetValue(val)
			local r, g, b = SmoothColour(val / #self.icons)
			self.bar:SetStatusBarColor(r, g, b)
			self.bar.bg:SetVertexColor(r, g, b, 0.3)

			icon.data[spellName] = copy(debuff)
			icon:Activate()
		end

		if (dose) then
			icon:ShowCount(debuff.stacks)
			icon.data[spellName].stacks = debuff.stacks
		end
	end
	
	-- OnDebuffRemoved
	function frame:OnDebuffRemoved(spellName, debuff, dose)
		--module:Print(format("frame:OnDebuffRemoved(%s)", spellName))

		local icon = self:GetIconForSpell(spellName)
--@debug@
		assert(icon)
--@no-debug@
		if (not icon.data) then
			return
		end

		local spellInfo = module.lookup[spellName]
--@debug@
		assert(spellInfo)
		assert(spellInfo.key == icon.key)
--@no-debug@

		if (spellInfo.exclusiveWith) then
			-- Special case for paladins
			if (module.classcount[spellInfo.class] < 2) then
				local altIcon = self:GetIconForKey(spellInfo.exclusiveWith)
				self.bar:SetMinMaxValues(0, self:GetMaxAvailable())
				altIcon:Deactivate()
			end
		end

		if (dose) then
			icon:ShowCount(debuff.stacks)
			icon.data[spellName].stacks = debuff.stacks
		else
			if (icon.data[spellName]) then
				local val = self.bar:GetValue() - 1	-- TEST: May need to validate against other spells in same class before decrementing
				self.bar:SetValue(val)
				local r, g, b = SmoothColour(val / #self.icons)
				self.bar:SetStatusBarColor(r, g, b)
				self.bar.bg:SetVertexColor(r, g, b, 0.3)

				icon:Deactivate()
				icon.data[spellName] = deepDel(icon.data[spellName])
			end
		end
	end

	function frame:UpdateBuffs()
		self:AllOff()

		local mob = engine:GetMobByUnit(self.unit)
		if (mob and mob.debuffs) then
			local on = 0
			for debuffName, debuff in pairs(mob.debuffs) do
				local icon = self:GetIconForSpell(debuffName)

				icon:Activate()
				on = on + 1

				if (not icon.data) then
					icon.data = new()
				end
				icon.data[debuffName] = copy(debuff)
			end

			self.bar:SetValue(on)

			local r, g, b = SmoothColour(on / #self.icons)
			self.bar:SetStatusBarColor(r, g, b)
			self.bar.bg:SetVertexColor(r, g, b, 0.3)
		end
	end

	function frame:SetTitle()
		if (self.backdrop) then
			if (UnitCanAttack("player", self.unit) and not UnitIsPlayer(self.unit)) then
				self.backdrop.title:SetText(UnitName(self.unit))
				self.backdrop.title:SetTextColor(module:ReactionColour(self.unit))
			else
				self.backdrop.title:SetText(module.label)
				self.backdrop.title:SetTextColor(1, 1, 1)
			end
		end
	end

	self.InitFrames = nil
end

-- ReactionColour
function Utopia:ReactionColour(unit)
	if (not UnitPlayerControlled(unit)) then
		if (UnitIsDeadOrGhost(unit)) then
			return 0.4, 0.4, 0.4
		elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) then
			return 0.6, 0.6, 0.6
		else
			local reaction = UnitReaction(unit, "player")
			if (reaction) then
				if (reaction >= 5) then
					return 0, 1, 0
				elseif (reaction <= 2) then
					return 1, 0, 0
				elseif (reaction == 3) then
					return 1, 0.5, 0
				else
					return 1, 1, 0
				end
			end
		end
	end
	return 0, 0.5, 1
end

-- SetFrameSize
function module:SetFrameSize()
	local frame = self.frame
	if (frame) then
		local db = db
		local cols = db.columns
		local rows = ceil(#frame.icons / cols)
		local size = db.size

		if (db.orientation == "H") then
			frame:SetWidth(size * cols + (db.hspacing * (cols - 1)))
			frame:SetHeight(size * rows + (db.vspacing * (rows - 1)) + db.barwidth)
		else
			frame:SetWidth(size * rows + (db.hspacing * (rows - 1)) + db.barwidth)
			frame:SetHeight(size * cols + (db.vspacing * (cols - 1)))
		end
	end
end

-- SetBarOrientation
function module:SetBarOrientation(info, n)
	if (n == "B" or n == "T") then
		db.side = n
	end

	local frame = self.frame
	if (frame) then
		self:SetFrameSize()

		frame.bar:ClearAllPoints()
		if (db.orientation == "H") then
			if (db.side == "B") then
				frame.bar:SetPoint("BOTTOMLEFT")
				frame.bar:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", 0, db.barwidth)
			else
				frame.bar:SetPoint("TOPLEFT")
				frame.bar:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 0, -db.barwidth)
			end
		else
			if (db.side == "B") then
				frame.bar:SetPoint("TOPRIGHT")
				frame.bar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", -db.barwidth, 0)
			else
				frame.bar:SetPoint("TOPLEFT")
				frame.bar:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", db.barwidth, 0)
			end
		end
	end
end

-- SetOrientation
function module:SetOrientation(info, newValue)
	local n
	if (newValue == "V" or newValue == "H") then
		db.orientation = newValue
	elseif (newValue == "T" or newValue == "B") then
		db.side = newValue
	end
	local n = db.orientation

	local frame = self.frame
	if (frame) then
		local prev, prevRow
		for i,icon in ipairs(frame.icons) do
			icon:ClearAllPoints()
			if (prev) then
				if ((i - 1) % db.columns == 0) then
					if (n == "H") then
						icon:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0, -db.vspacing)
					else
						icon:SetPoint("TOPLEFT", prevRow, "TOPRIGHT", db.hspacing, 0)
					end
					prevRow = icon
				else
					if (n == "H") then
						icon:SetPoint("TOPLEFT", prev, "TOPRIGHT", db.hspacing, 0)
					else
						icon:SetPoint("TOPLEFT", prev, "BOTTOMLEFT", 0, -db.vspacing)
					end
				end
			else
				if (db.side == "T") then
					if (n == "H") then
						icon:SetPoint("TOPLEFT", 0, -db.barwidth)
					else
						icon:SetPoint("TOPLEFT", db.barwidth, 0)
					end
				else
					icon:SetPoint("TOPLEFT")
				end
				prevRow = icon
			end
			prev = icon
		end
		self:SetBarOrientation()
		self:UpdateBackdrop()
	end
end

-- UpdateBackdrop
local bgFrame = {insets = {}}
function module:UpdateBackdrop()
	if (db.Background.Enable) then
		bgFrame.bgFile = SM:Fetch("background", db.Background.Texture)
		bgFrame.edgeFile = SM:Fetch("border", db.Background.BorderTexture)
		bgFrame.tile = db.Background.Tile
		bgFrame.tileSize = db.Background.TileSize
		bgFrame.edgeSize = db.Background.EdgeSize
		local inset = floor(db.Background.EdgeSize / 4)
		bgFrame.insets.left = inset
		bgFrame.insets.right = inset
		bgFrame.insets.top = inset
		bgFrame.insets.bottom = inset
	
		local bd = self.frame.backdrop
		if (not bd) then
			-- Backdrop is going to be larger than the frame proper
			local bd = CreateFrame("Frame", nil, self.frame)
			self.frame.backdrop = bd
			self.frame:SetFrameLevel(bd:GetFrameLevel() + 2)
			bd:SetScript("OnDragStart", dragStart)
			bd:SetScript("OnDragStop", dragStop)
			bd:RegisterForDrag("LeftButton")
			if (not db.locked) then
				bd:EnableMouse(true)
			end

			bd.title = bd:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			bd.title:SetText(self.label)
			bd.title:SetPoint("BOTTOMLEFT", self.frame, "TOPLEFT", -1, 1)
			bd.title:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT", 1, 11)
			bd.title:SetJustifyH("LEFT")
		end

		local esd2 = db.Background.EdgeSize / 2
		self.frame.backdrop:ClearAllPoints()
		self.frame.backdrop:SetPoint("TOPLEFT", -esd2, esd2)
		self.frame.backdrop:SetPoint("BOTTOMRIGHT", esd2, -esd2)
	
		self.frame.backdrop:SetBackdrop(bgFrame)
	
		local c = db.Background.Color
		self.frame.backdrop:SetBackdropColor(c.r, c.g, c.b, c.a)
	
		c = db.Background.BorderColor
		self.frame.backdrop:SetBackdropBorderColor(c.r, c.g, c.b, c.a)

		self.frame.backdrop:Show()
	else
		if (self.frame.backdrop) then
			self.frame.backdrop:Hide()
		end
	end
end

-- ToggleActive
function module:ToggleActive()
	db.enable = not db.enable
	self:ShowWhen()
end

-- SetIconSize
function module:SetIconSize(info, n)
	db.size = n
	
	if (self.frame) then
		for i,icon in pairs(self.frame.icons) do
			icon:SetWidth(n)
			icon:SetHeight(n)
		end

		self:SetFrameSize()
		self:SavePosition(module.frame)
	end
end

-- SetScale
function module:SetScale()
	if (self.frame) then
		self.frame:SetScale(db.scale)
	end
end

-- ShowWhen
function module:ShowWhen()
	if (db.enable) then
		if (not self.frame) then
			self:InitFrames()
		end

		local show = db.show.always
		if (not show) then
			if (UnitCanAttack("player", self.frame.unit) and not UnitIsDead(self.frame.unit)) then
				local pvpInstance = select(2, IsInInstance()) == "pvp"
				if (db.show.boss) then
					if (UnitClassification(self.frame.unit) == "worldboss") then
						show = true
					end
				else
					if (db.show.raid) then
						show = GetNumRaidMembers() > 0 and not pvpInstance
					end
					if (not show and db.show.party) then
						show = GetNumPartyMembers() > 0 and not pvpInstance
					end
					if (not show and db.show.solo) then
						show = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0
					end
				end
			end
		end

		if (show) then
			self.frame:Show()
			self.frame:SetTitle()
		else
			self.frame:Hide()
		end
	elseif (self.frame) then
		self.frame:Hide()
	end
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	self:CheckForMissingTalents()
	self:ShowWhen()
end

-- PARTY_MEMBER_ENABLE
function module:PARTY_MEMBER_ENABLE()
	self:CheckForMissingTalents()
end

-- frameOnUpdate
local function frameOnUpdate(self, elapsed)
	if (module.rosterUpdate) then
		module:OnRaidRosterUpdate()
	end
	module.frame:SetScript("OnUpdate", nil)
end

-- PARTY_MEMBERS_CHANGED
function module:PARTY_MEMBERS_CHANGED()
	self.rosterUpdate = true
	self.frame:SetScript("OnUpdate", frameOnUpdate)
end

-- OnRaidRosterUpdate
function module:OnRaidRosterUpdate()
	self.distribution = nil
	if (select(2, IsInInstance()) ~= "pvp") then
		if (GetNumRaidMembers() > 0) then
			self.distribution = "RAID"
		elseif (GetNumPartyMembers() > 0) then
			self.distribution = "PARTY"
		end
	end

	if (self.distribution) then
		self:SendCommMessage(commPrefix, "HELLO "..self.version, self.distribution)
	end

	self:ScanRoster()
	if (self.distribution or db.show.always) then
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
	else
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	end

	self:PLAYER_TARGET_CHANGED()
end

do
	local function iter(t)
		local key = t.id
		if (t.mode == "raid") then
			if (key > t.r) then
				del(t)
				return nil
			end
			ret = "raid"..key
		else
			if (key > t.p) then
				del(t)
				return nil
			end
			ret = key == 0 and "player" or "party"..key
		end
		t.id = key + 1
		return ret
	end

	-- IterateRoster
	function module:IterateRoster()
		local t = new()
		if (GetNumRaidMembers() > 0) then
			t.mode = "raid"
			t.id = 1
			t.r = GetNumRaidMembers()
		else
			t.mode = "party"
			t.id = 0
			t.p = GetNumPartyMembers()
		end
		return iter, t
	end
end

-- ScanRoster
function module:ScanRoster()
	del(self.potential)
	self.potential = new()
	del(self.spellPotential)
	self.spellPotential = new()
	del(self.petPotential)
	self.petPotential = new()
	del(self.pets)
	self.pets = new()

	local oldTalents
	if (self.talents) then
		oldTalents = new()
		for name,talents in pairs(self.talents) do
			oldTalents[name] = true
		end
	end

	self.classcount = setmetatable({}, {__index = function() return 0 end})
	for unit in self:IterateRoster() do
		if (oldTalents) then
			local name = UnitFullName(unit)
			oldTalents[name] = nil
		end
		
		local _, class = UnitClass(unit)
		if (class) then
			self.classcount[class] = self.classcount[class] + 1

			local talents = self:GetTalents(unit)

			self:AddPotential(unit, talents)

			if (class == "HUNTER") then
				self:UpdateUnitPet(unit, true)
			end
		end
	end

	if (oldTalents) then
		for name in pairs(oldTalents) do
			self.talents[name] = del(self.talents[name])
		end
		del(oldTalents)
	end

	self:MakePetPotentialKeys()
	if (self.frame) then
		self.frame:UpdateBuffs()
	end
end

-- UNIT_PET
function module:UNIT_PET(event, unit)
	self:UpdateUnitPet(unit)
end

-- UpdateUnitPet
function module:UpdateUnitPet(unit, noPetUpdate)
	local owner = UnitFullName(unit)
	--self:Print(format("UpdateUnitPet(%s) - %s", tostring(unit), tostring(owner)))
	local petid
	if (unit == "player") then
		petid = "pet"
	else
		petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
	end

	local cType, remove
	if (UnitExists(petid) and UnitCreatureFamily(petid)) then
		if (not self.pets) then
			self.pets = new()
		end
		cType = UnitCreatureFamily(petid)
		--self:Print(format("%s now has a %s pet", owner or "nil", cType or "nil"))
		if (self.pets[owner] == cType) then
			return				-- No change
		end

		self.pets[owner] = cType
	else
		if (self.pets) then
			cType = self.pets[owner]
			--self:Print(format("%s lost his %s pet", owner or "nil", cType or "nil"))

			if (not self.pets[owner]) then
				return				-- No change
			end

			remove = true
			self.pets[owner] = nil
			if (not next(self.pets)) then
				self.pets = del(self.pets)
			end
		end
	end

	if (cType) then
		local c = self.classes.HUNTER
		if (c) then
			for key, name in pairs(c) do
				local info = self.lookup[name]

				if (info.pet == cType) then
					local list = self.petPotential[name]

					if (remove) then
						if (list and list[owner]) then
							--self:Print(format("Removing |cFF80FF80%s|r from potential pet skills", name))
							list[owner] = nil
							if (not next(list)) then
								self.petPotential[name] = del(list)
								change = true
							end
						end
					else
						if (not list) then
							list = new()
							self.petPotential[name] = list
						end
						--self:Print(format("Adding |cFF80FF80%s|r from potential pet skills", name))
						if (not list[owner]) then
							change = true
							list[owner] = true
						end
					end
				end
			end

			if (change) then
				if (not noPetUpdate) then
					self:MakePetPotentialKeys()
					self.frame:UpdateBuffs()
				end
			end
		end
	end
end

-- MakePetPotentialKeys
function module:MakePetPotentialKeys()
	del(self.petPotentialKeys)
	self.petPotentialKeys = new()

	if (self.petPotential) then
		for spellName,list in pairs(self.petPotential) do
			local spell = self.lookup[spellName]
			assert(spell, "spell")

			self.petPotentialKeys[spell.key] = true
		end
	end
end

-- AddPotential
function module:AddPotential(unit, talents)
	local _, class = UnitClass(unit)
	if (class) then
		for category, spells in pairs(self.data) do
			for spellName, info in pairs(spells.spells) do
				if (info.class == class) then
					local needsPet = class == "HUNTER" and info.pet

					if (not info.requiredTalent and not needsPet) then
						self.potential[info.key] = 0
						self.spellPotential[spellName] = 0
					else
						if (talents) then
							local level = self:HasTalent(unit, info.requiredTalent)
							if (level and level > 0) then
--@debug@
								if (info.maxTalentPoints and level > info.maxTalentPoints) then
									error(format("%s's talent points(%d) for %s is greater than maxTalentPoints(%d)", UnitName(unit), level, info.maxTalentPoints))
								end
--@no-debug@
								if (not needsPet) then
									self.potential[info.key] = max(self.potential[info.key] or 0, level)
									self.spellPotential[spellName] = max(self.spellPotential[spellName] or 0, level)
								end
							end
						end
					end
				end
			end
		end
	end
end

-- HasTalent
function module:HasTalent(unit, required)
	if (self.classTalentData) then
		local _, class = UnitClass(unit)
		if (class) then
			local data = self.classTalentData[class]
			if (not data) then
				if (class == select(2,UnitClass("player"))) then
					self:GetClassTalentData("player")
				elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
					if (UnitIsVisible(unit) and CanInspect(unit)) then
						if (not UnitIsPlayer(unit)) then
							NotifyInspect(unit)
						end
						self:GetClassTalentData(unit)
					end
				end
				data = self.classTalentData[class]
			end

			if (data) then
				local info = data.list[required]
				if (info) then
					local index = info.index
					local tree = info.treeIndex
	
					local talents = self.talents and self.talents[UnitFullName(unit)]
					if (talents) then
						local str = talents[tree]
						if (str) then
							local amount = tonumber(str:sub(index, index))
							return amount > 0 and amount or nil
						end
					end
				end
			end
		end
	end
end

-- OnReceiveTalents
function module:OnReceiveTalents(fullname, talents)
	if (talents and next(talents) and type(talents[1]) == "string") then
		if (self.missingTalents) then
			if (self.missingTalents[fullname]) then
				self.missingTalents[fullname] = nil
				engine:ScanRoster()	-- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log
			end
		end

		if (self.frame) then
			self:AddPotential(fullname, talents)
			self.frame:UpdateBuffs()
		end
	end
end

-- GetTalents
do
	local RockComm = LibStub("LibRockComm-1.0", true)
	if (RockComm) then
		local function talentReceptor(sender, data)
			if (not module.talents) then
				module.talents = new()
			end

			if (type(data) == "table" and #data == 3 and type(data[1]) == "table") then
				local n = new()
				
				for i = 1,3 do
					local str = ""
					for j = 1,#data[i] do
						str = str .. tostring(data[i][j])
					end
					tinsert(n, str)
				end
				module.talents[sender] = n
				module:OnReceiveTalents(sender, n)
			end
		end
		RockComm:AddTalentReceptor(talentReceptor)
	end
	
	function module:GetTalents(unit)
		local name = UnitFullName(unit)
		local talents = self.talents and self.talents[name]

		if (not talents and not self.talentRecursion) then
			--self:Print("Want talents for "..name)
			if (not self.missingTalents) then
				self.missingTalents = new()
			end
			--self:Print("Adding "..name.." to missingTalents")
			self.missingTalents[name] = true

			if (UnitIsUnit("player", unit)) then
				talents = self:TalentQuery_Ready(nil, UnitName(unit))
			elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
				if (UnitIsVisible(unit) and CanInspect(unit)) then
					--self:Print("Querying talents via INSPECT for "..name)
					TalentQuery:Query(unit)

				elseif (self.users and self.users[name]) then
					if (not self.commQueried or not self.commQueried[name]) then
						--self:Print("Querying talents via Utopia comms for "..name)
						if (not self.commQueried) then
							self.commQueried = new()
						end
						self.commQueried[name] = true
						self:SendCommMessage(commPrefix, "REQUESTTALENTS", "WHISPER", name)
					end

				elseif (RockComm) then
					if (not self.rockCommQueried or not self.rockCommQueried[name]) then
						--self:Print("Querying talents via RockComm for "..name)
						if (not self.rockCommQueried) then
							self.rockCommQueried = new()
						end
						self.rockCommQueried[name] = true
						RockComm:QueryTalents("WHISPER", name)
					end
				end
			end
		end

		return talents
	end
end

-- CheckForMissingTalents
function module:CheckForMissingTalents()
	if (self.missingTalents and self.loggedIn and UnitInParty("player")) then
		for name in pairs(self.missingTalents) do
			-- We can use IsVisible, becuase we've already tried once to request
			-- talents via comms and failed, So we need them close to inspect
			if (UnitInRaid(name) or UnitInParty(name)) then
				if (CanInspect(name)) then
					--self:Print("CheckForMissingTalents - "..name)
					self:GetTalents(name)
				end
			else
				if (name ~= "player" and name ~= UnitName("player")) then
					--self:Print("CheckForMissingTalents - removing "..name.." (not in party/raid)")
					self.missingTalents[name] = nil
				end
			end
		end
	end
end

-- GetClassTalentData
function module:GetClassTalentData(unit)
	local _, class = UnitClass(unit)

	if (class) then
		local data = self.classTalentData and self.classTalentData[class]
		if (not data) then
			local isnotplayer = not UnitIsUnit("player", unit)
			if (GetNumTalentTabs(isnotplayer) > 0) then
				data = new()
				data.list = new()

				for tab = 1, GetNumTalentTabs(isnotplayer) do
					local tree = new()
					local pointssptent
					tree.name, tree.icon, pointssptent = GetTalentTabInfo(tab, isnotplayer)
					tinsert(data, tree)

					tree.list = new()
					for i = 1,GetNumTalents(tab, isnotplayer) do
						local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer)

						local entry = new()
						entry.name = name
						entry.icon = icon
						entry.tier = tier
						entry.column = column
						entry.maxRank = maxRank
						entry.index = i
						entry.treeIndex = tab
						tinsert(tree.list, entry)
						data.list[name] = entry
					end
				end

				if (next(data)) then
					if (not self.classTalentData) then
						self.classTalentData = new()
					end
					self.classTalentData[class] = data
				end
			end
		end
	end
end

-- TalentQuery_Ready
function module:TalentQuery_Ready(e, name, realm)
	local namerealm = realm and realm ~= "" and name .. "-" .. realm or name
	local isnotplayer = (namerealm ~= UnitName("player"))

	self:GetClassTalentData(namerealm)

	if (GetNumTalentTabs(isnotplayer) and GetNumTalentTabs(isnotplayer) >= 3 and GetNumTalents(1, isnotplayer) > 0) then
		local n = new()
		for tab = 1, GetNumTalentTabs(isnotplayer) do
			local str
			for i = 1,GetNumTalents(tab, isnotplayer) do
				local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer)
				str = (str or "") .. tostring(currentRank)
			end
			tinsert(n, str)
		end

		if (not self.talents) then
			self.talents = new()
		end
		self.talents[namerealm] = n
		self:OnReceiveTalents(namerealm, n)

		return n
	end
end

-- PLAYER_ENTERING_WORLD
function module:PLAYER_ENTERING_WORLD()
	self.loggedIn = true

	if (not self.initialLogin) then
		self.initialLogin = true
		self:ScheduleTimer("PLAYER_ENTERING_WORLD", 2, self)
	end
	
	self:CheckForMissingTalents()

	if (select(2, IsInInstance()) ~= "pvp") then
		engine:Start()
	end
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self.loggedIn = nil
	engine:Stop()
end

-- OnCommReceived
function module:OnCommReceived(prefix, msg, channel, sender)
	if (sender == playerName) then
		return
	end

	if (prefix == commPrefix) then
		local cmd, str = msg:match("^(%a+) *(.*)$")
		if (not cmd) then
			return
		end

		if (cmd == "TALENTS") then
			local list = new(strsplit(",", str))
			if (next(list)) then
				self.talents[sender] = list
				self:OnReceiveTalents(sender, list)
			end
			del(list)

		elseif (cmd == "REQUESTTALENTS") then
			local t = self.talents[UnitName("player")]
			if (t) then
				self:SendCommMessage(commPrefix, format("TALENTS %s,%s,%s", t[1], t[2], t[3]), "WHISPER", sender)
			end

		elseif (cmd == "HELLO") then
			if (channel ~= "WHISPER" and sender ~= UnitName("player")) then
				if (not self.users) then
					self.users = {}
				end
				self.users[sender] = tonumber(str) or str
				if (str ~= "ZOMG") then
					self:SendCommMessage(commPrefix, "HELLO "..self.version, "WHISPER", sender)
				end
			end
		end
	end
end

-- OpenConfig
function module:OpenConfig(input)
	if (input == "test") then
		for class,data in pairs(self.classTalentData) do
			if (not RAID_CLASS_COLORS[class]) then
				self:Print(format("Bad class '%s' in classTalentData", tostring(class)))
			elseif (not data.list or not data[1] or not data[2] or not data[3]) then
				self:Print(format("Bad data for '%s' in classTalentData", tostring(class)))
			end
		end
		for name,data in pairs(self.talents) do
			if (type(data) ~= "table") then
				self:Print(format("Invalid talent info for %s", tostring(name)))
			elseif (#data < 3) then
				self:Print(format("Incomplete talent info for %s", tostring(name)))
			elseif (type(data[1]) ~= "string" or type(data[2]) ~= "string" or type(data[3]) ~= "string") then
				self:Print(format("Talent info is not strings for %s", tostring(name)))
			end
		end

	elseif (input == "config" or not InterfaceOptionsFrame:IsResizable()) then
		--module.options.args.aceConfig.guiHidden = true
		InterfaceOptionsFrame:Hide()
		AceConfigDialog:SetDefaultSize("Utopia", 650, 500)
		AceConfigDialog:Open("Utopia")
	else
		--module.options.args.aceConfig.guiHidden = false
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.Utopia)
	end
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
	self:SetOrientation()
	self:ShowWhen()
end

-- SetupOptions
function module:SetupOptions()
	self.optionsFrames = {}

	-- setup options table
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Utopia", self.options)
	local ACD3 = LibStub("AceConfigDialog-3.0")

	self.optionsFrames.Utopia = ACD3:AddToBlizOptions("Utopia", nil, nil, "General")
	self.optionsFrames.ShowWhen = ACD3:AddToBlizOptions("Utopia", L["Show When"], "Utopia", "show")
	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), L["Profiles"])

	-- Add ordering data to the option table generated by AceDBOptions-3.0
	self.options.args.Profiles.order = -2

	self:RegisterChatCommand("utopia", "OpenConfig")
end

-- RegisterModuleOptions
function module:RegisterModuleOptions(name, optionTbl, displayName)
	self.options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Utopia", displayName, "Utopia", name)
end

-- OnModuleInitialize
function module:OnInitialize()
	new, del, copy, deepDel = self.new, self.del, self.copy, self.deepDel
	Gradient, UnitFullName, ClassColour, SmoothColour = self.Gradient, self.UnitFullName, self.ClassColour, self.SmoothColour
	rotate = Utopia.rotate
	engine = UtopiaDebuffActivity

	self.label = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
	self.title = Gradient(L["TITLE"].." r"..module.version, 1, 0.1, 0.1, 1, 1, 0.1)

	if (not engine) then
		error("Utopia engine not loaded (Did you restart WoW after installing?)")
	end

	module.options.name = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
	TalentQuery.RegisterCallback(self, "TalentQuery_Ready")
	playerClass = select(2, UnitClass("player"))

	SM:Register("border", "Vista", [[Interface\Addons\Utopia\Textures\Border-Vista]])

	self.db = LibStub("AceDB-3.0"):New("UtopiaDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

	self.db:RegisterDefaults({
		profile = {
			enable = true,
			show = {
				always = false,
				solo = false,
				party = true,
				raid = true,
			},
			size = 20,
			scale = 1,
			texture = "Blizzard",
			orientation = "H",
			side = "B",
			barwidth = 7,
			hspacing = 0,
			vspacing = 0,
			columns = 15,
			colour = {
				active = {r = 1, g = 1, b = 1, a = 1},
				inactive = {r = 1, g = 0, b = 0, a = 1},
				unavailable = {r = 1, g = 1, b = 1, a = 0.25},
			},
			Background = {
				Enable = true,
				Texture = "Solid",
				BorderTexture = "Blizzard Tooltip",
				Color = {r = 0.06, g = 0.06, b = 0.06, a = 0.5},
				BorderColor = {r = 1, g = 1, b = 1, a = 1},
				Tile = false,
				TileSize = 32,
				EdgeSize = 9,
			},
		}
	})
	db = self.db.profile

	self:SetupOptions()
	self:RegisterComm(commPrefix)

	self.OnInitialize = nil
end

-- OnModuleEnable
function module:OnEnable()
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		self:InitData()
		self:OnRaidRosterUpdate()
		self:PLAYER_TARGET_CHANGED()

		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("PARTY_MEMBER_ENABLE")
		self:RegisterEvent("UNIT_PET")
		self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
	end
end

-- OnDisable
function module:OnDisable()
	engine:Stop()
	self.lookup = del(self.lookup)
	self.mappingLookup = del(self.mappingLookup)
	self.classes = deepDel(self.classes)
	self.frame:Hide()
end
